#' Neighbors Forecaster
#'
#' Predict next serie as a weighted combination of "futures of the past" days,
#' where days in the past are chosen and weighted according to some similarity measures.
#'
#' The main method is \code{predictShape()}, taking arguments data, today, memory,
#' predict_from, horizon respectively for the dataset (object output of
#' \code{getData()}), the current index, the data depth (in days), the first predicted
#' hour and the last predicted hour.
#' In addition, optional arguments can be passed:
#' \itemize{
#'   \item local : TRUE (default) to constrain neighbors to be "same days within same
#'     season"
#'   \item simtype : 'endo' for a similarity based on the series only,<cr>
#'             'exo' for a similarity based on exogenous variables only,<cr>
#'             'mix' for the product of 'endo' and 'exo',<cr>
#'             'none' (default) to apply a simple average: no computed weights
#'   \item window : A window for similarities computations; override cross-validation
#'     window estimation.
#' }
#' The method is summarized as follows:
#' \enumerate{
#'   \item Determine N (=20) recent days without missing values, and followed by a
#'     tomorrow also without missing values.
#'   \item Optimize the window parameters (if relevant) on the N chosen days.
#'   \item Considering the optimized window, compute the neighbors (with locality
#'     constraint or not), compute their similarities -- using a gaussian kernel if
#'     simtype != "none" -- and average accordingly the "tomorrows of neigbors" to
#'     obtain the final prediction.
#' }
#'
#' @usage # NeighborsForecaster$new(pjump)
#'
#' @docType class
#' @format R6 class, inherits Forecaster
#' @aliases F_Neighbors
#'
NeighborsForecaster = R6::R6Class("NeighborsForecaster",
	inherit = Forecaster,

	public = list(
		predictShape = function(data, today, memory, predict_from, horizon, ...)
		{
			# (re)initialize computed parameters
			private$.params <- list("weights"=NA, "indices"=NA, "window"=NA)

			# Do not forecast on days with NAs (TODO: softer condition...)
			if (any(is.na(data$getSerie(today-1))) ||
				(predict_from>=2 && any(is.na(data$getSerie(today)[1:(predict_from-1)]))))
			{
				return (NA)
			}

			# Determine indices of no-NAs days preceded by no-NAs yerstedays
			tdays = .getNoNA2(data, max(today-memory,2), today-1)

			# Get optional args
			local = ifelse(hasArg("local"), list(...)$local, TRUE) #same level + season?
			simtype = ifelse(hasArg("simtype"), list(...)$simtype, "none") #or "endo", or "exo"
			if (hasArg("window"))
			{
				return ( private$.predictShapeAux(data,
					tdays, today, predict_from, horizon, local, list(...)$window, simtype, TRUE) )
			}

			# Indices of similar days for cross-validation; TODO: 20 = magic number
			cv_days = getSimilarDaysIndices(today, data, limit=20, same_season=FALSE,
				days_in=tdays)

			# Optimize h : h |--> sum of prediction errors on last N "similar" days
			errorOnLastNdays = function(window, simtype)
			{
				error = 0
				nb_jours = 0
				for (i in seq_along(cv_days))
				{
					# mix_strategy is never used here (simtype != "mix"), therefore left blank
					prediction = private$.predictShapeAux(data, tdays, cv_days[i], predict_from,
						horizon, local, window, simtype, FALSE)
					if (!is.na(prediction[1]))
					{
						nb_jours = nb_jours + 1
						error = error +
							mean((data$getSerie(cv_days[i])[predict_from:horizon] - prediction)^2)
					}
				}
				return (error / nb_jours)
			}

			# TODO: 7 == magic number
			if (simtype=="endo" || simtype=="mix")
			{
				best_window_endo = optimize(
					errorOnLastNdays, c(0,7), simtype="endo")$minimum
			}
			if (simtype=="exo" || simtype=="mix")
			{
				best_window_exo = optimize(
					errorOnLastNdays, c(0,7), simtype="exo")$minimum
			}

			best_window =
				if (simtype == "endo")
					best_window_endo
				else if (simtype == "exo")
					best_window_exo
				else if (simtype == "mix")
					c(best_window_endo,best_window_exo)
				else #none: value doesn't matter
					1

			return( private$.predictShapeAux(data, tdays, today, predict_from, horizon, local,
				best_window, simtype, TRUE) )
		}
	),
	private = list(
		# Precondition: "today" is full (no NAs)
		.predictShapeAux = function(data, tdays, today, predict_from, horizon, local, window,
			simtype, final_call)
		{
			tdays_cut = tdays[ tdays <= today-1 ]
			if (length(tdays_cut) <= 1)
				return (NA)

			if (local)
			{
				# TODO: 60 == magic number
				tdays = getSimilarDaysIndices(today, data, limit=60, same_season=TRUE,
					days_in=tdays_cut)
				if (length(tdays) <= 1)
					return (NA)
				# TODO: 10, 12 == magic numbers
				tdays = .getConstrainedNeighbs(today,data,tdays,min_neighbs=10,max_neighbs=12)
				if (length(tdays) == 1)
				{
					if (final_call)
					{
						private$.params$weights <- 1
						private$.params$indices <- tdays
						private$.params$window <- 1
					}
					return ( data$getSerie(tdays[1])[predict_from:horizon] )
				}
			}
			else
				tdays = tdays_cut #no conditioning

			if (simtype == "endo" || simtype == "mix")
			{
				# Compute endogen similarities using given window
				window_endo = ifelse(simtype=="mix", window[1], window)

				# Distances from last observed day to days in the past
				lastSerie = c( data$getSerie(today-1),
					data$getSerie(today)[if (predict_from>=2) 1:(predict_from-1) else c()] )
				distances2 = sapply(tdays, function(i) {
					delta = lastSerie - c(data$getSerie(i-1),
						data$getSerie(i)[if (predict_from>=2) 1:(predict_from-1) else c()])
					sqrt(mean(delta^2))
				})

				simils_endo <- .computeSimils(distances2, window_endo)
			}

			if (simtype == "exo" || simtype == "mix")
			{
				# Compute exogen similarities using given window
				window_exo = ifelse(simtype=="mix", window[2], window)

				M = matrix( ncol=1+length(tdays), nrow=1+length(data$getExo(1)) )
				M[,1] = c( data$getLevelHat(today), as.double(data$getExoHat(today)) )
				for (i in seq_along(tdays))
					M[,i+1] = c( data$getLevel(tdays[i]), as.double(data$getExo(tdays[i])) )

				sigma = cov(t(M)) #NOTE: robust covariance is way too slow
				# TODO: 10 == magic number; more robust way == det, or always ginv()
				sigma_inv =
					if (length(tdays) > 10)
						solve(sigma)
					else
						MASS::ginv(sigma)

				# Distances from last observed day to days in the past
				distances2 = sapply(seq_along(tdays), function(i) {
					delta = M[,1] - M[,i+1]
					delta %*% sigma_inv %*% delta
				})

				simils_exo <- .computeSimils(distances2, window_exo)
			}

			similarities =
				if (simtype == "exo")
					simils_exo
				else if (simtype == "endo")
					simils_endo
				else if (simtype == "mix")
					simils_endo * simils_exo
				else #none
					rep(1, length(tdays))
			similarities = similarities / sum(similarities)

			prediction = rep(0, horizon-predict_from+1)
			for (i in seq_along(tdays))
			{
				prediction = prediction +
					similarities[i] * data$getSerie(tdays[i])[predict_from:horizon]
			}

			if (final_call)
			{
				private$.params$weights <- similarities
				private$.params$indices <- tdays
				private$.params$window <-
					if (simtype=="endo")
						window_endo
					else if (simtype=="exo")
						window_exo
					else if (simtype=="mix")
						c(window_endo,window_exo)
					else #none
						1
			}

			return (prediction)
		}
	)
)

# getConstrainedNeighbs
#
# Get indices of neighbors of similar pollution level (among same season + day type).
#
# @param today Index of current day
# @param data Object of class Data
# @param tdays Current set of "second days" (no-NA pairs)
# @param min_neighbs Minimum number of points in a neighborhood
# @param max_neighbs Maximum number of points in a neighborhood
#
.getConstrainedNeighbs = function(today, data, tdays, min_neighbs=10, max_neighbs=12)
{
	levelToday = data$getLevelHat(today)
	levelYersteday = data$getLevel(today-1)
	distances = sapply(tdays, function(i) {
		sqrt((data$getLevel(i-1)-levelYersteday)^2 + (data$getLevel(i)-levelToday)^2)
	})
	#TODO: 1, +1, +3 : magic numbers
	dist_thresh = 1
	min_neighbs = min(min_neighbs,length(tdays))
	repeat
	{
		same_pollution = (distances <= dist_thresh)
		nb_neighbs = sum(same_pollution)
		if (nb_neighbs >= min_neighbs) #will eventually happen
			break
		dist_thresh = dist_thresh + ifelse(dist_thresh>1,3,1)
	}
	tdays = tdays[same_pollution]
	max_neighbs = 12
	if (nb_neighbs > max_neighbs)
	{
		# Keep only max_neighbs closest neighbors
		tdays = tdays[ order(distances[same_pollution])[1:max_neighbs] ]
	}
	tdays
}

# compute similarities
#
# Apply the gaussian kernel on computed squared distances.
#
# @param distances2 Squared distances
# @param window Window parameter for the kernel
#
.computeSimils <- function(distances2, window)
{
	sd_dist = sd(distances2)
	if (sd_dist < .25 * sqrt(.Machine$double.eps))
	{
#		warning("All computed distances are very close: stdev too small")
		sd_dist = 1 #mostly for tests... FIXME:
	}
	exp(-distances2/(sd_dist*window^2))
}
