#' Compute forecast
#'
#' Predict time-series curves ("tomorrows") at the selected days indices ("todays").
#' This function just runs a loop over all requested indices, and stores the individual
#' forecasts into a list which is then turned into a Forecast object.
#'
#' @param data Object of class Data, output of \code{getData()}.
#' @param indices Indices where to forecast (the day after); integers relative to the
#'   beginning of data, or (convertible to) Date objects.
#' @param forecaster Name of the main forecaster; more details: ?F_<forecastername>
#' \itemize{
#'   \item Persistence : use last (similar, next) day
#'   \item Neighbors : weighted tomorrows of similar days
#'   \item Average : average tomorrow of all same day-in-week
#'   \item Zero : just output 0 (benchmarking purpose)
#' }
#' @param pjump Function to predict the jump at the interface between two days;
#'   more details: ?J_<functionname>
#' \itemize{
#'   \item Persistence : use last (similar, next) day
#'   \item Neighbors: re-use the weights from F_Neighbors
#'   \item Zero: just output 0 (no adjustment)
#' }
#' @param memory Data depth (in days) to be used for prediction.
#' @param horizon Number of time steps to predict.
#' @param ncores Number of cores for parallel execution (1 to disable).
#' @param ... Additional parameters for the forecasting models.
#'
#' @return An object of class Forecast
#'
#' @examples
#' ts_data <- system.file("extdata","pm10_mesures_H_loc.csv",package="talweg")
#' exo_data <- system.file("extdata","meteo_extra_noNAs.csv",package="talweg")
#' data <- getData(ts_data, exo_data, input_tz="GMT", working_tz="GMT",
#'   predict_at=7, limit=200)
#' pred <- computeForecast(data, 100:130, "Persistence", "Zero",
#'   memory=50, horizon=12, ncores=1)
#' \dontrun{#Sketch for real-time mode:
#' data <- Data$new()
#' forecaster <- MyForecaster$new(myJumpPredictFunc)
#' repeat {
#'   # In the morning 7am+ or afternoon 1pm+:
#'   data$append(
#'     times_from_H+1_yersteday_to_Hnow,
#'     PM10_values_of_last_24h,
#'     exogenous_measures_of_last_24h,
#'     exogenous_predictions_for_next_24h)
#'   pred <- forecaster$predictSerie(data, data$getSize(), ...)
#'   #do_something_with_pred
#' }}
#' @export
computeForecast = function(data, indices, forecaster, pjump, predict_from,
	memory=Inf, horizon=length(data$getSerie(1)), ncores=3, ...)
{
	# (basic) Arguments sanity checks
	predict_from = as.integer(predict_from)[1]
	if (! predict_from %in% 1:length(data$getSerie(1)))
		stop("predict_from in [1,24] (hours)")
	horizon = as.integer(horizon)[1]
	if (horizon<=predict_from || horizon>length(data$getSerie(1)))
		stop("Horizon in [predict_from+1,24] (hours)")
	integer_indices = sapply(indices, function(i) dateIndexToInteger(i,data))
	if (any(integer_indices<=0 | integer_indices>data$getSize()))
		stop("Indices out of range")
	if (!is.character(forecaster) || !is.character(pjump))
		stop("forecaster (name) and pjump (function) should be of class character")

	pred = Forecast$new( sapply(indices, function(i) integerIndexToDate(i,data)) )
	forecaster_class_name = getFromNamespace(
		paste(forecaster,"Forecaster",sep=""), "talweg")
	forecaster = forecaster_class_name$new( #.pjump =
		getFromNamespace(paste("get",pjump,"JumpPredict",sep=""), "talweg"))

	computeOneForecast <- function(i)
	{
		list(
			"forecast" = forecaster$predictSerie(data,i,memory,predict_from,horizon,...),
			"params" = forecaster$getParameters(),
			"index" = i )
	}

	p <-
		if (ncores > 1 && requireNamespace("parallel",quietly=TRUE))
			parallel::mclapply(integer_indices, computeOneForecast, mc.cores=ncores)
		else
			lapply(integer_indices, computeOneForecast)

	# TODO: find a way to fill pred in //...
	for (i in seq_along(integer_indices))
	{
		pred$append(
			forecast = p[[i]]$forecast,
			params = p[[i]]$params,
			index_in_data = p[[i]]$index
		)
	}
	pred
}
